#!/usr/bin/env python

#***************************************************************************
#                        libcdemu.py  -  description
#                             -------------------
#    copyright            : (C) 2003 by Robert Penz
#    email                : robert.penz@outertech.com
#***************************************************************************

#***************************************************************************
#*                                                                         *
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU General Public License as published by  *
#*   the Free Software Foundation; either version 2 of the License, or     *
#*   (at your option) any later version.                                   *
#*                                                                         *
#***************************************************************************

import sys
import os
import re
import string
import struct
import fcntl
import stat

# this const is used by get_device_from_drive_number
# which adds the drive number at the end of this string
device_path="/dev/cdemu"

CUE_TRACK   = 0
CUE_MODE    = 1
CUE_INDEXES = 2

ENTRY_MODE  = 0
ENTRY_MIN   = 1
ENTRY_SEC   = 2
ENTRY_FRAME = 3
ENTRY_STARTSECTOR = 4
ENTRY_STARTOFFSET = 5

CDEMU_LOADCD = 0x5400   # load the virtual cd
CDEMU_UNLOADCD = 0x5401 # unload the virtual cd

UNKNOWN =-1
AUDIO = 0
MODE1 = 1
MODE2 = 2
MODE1_2352 = 10
MODE2_2352 = 20
MODE1_2048 = 30
MODE2_2336 = 40
LEADOUT = 100

# Some generally useful CD-ROM information -- copy and past from include/linux/cdrom.h
CD_MINS        =   74 #/* max. minutes per CD, not really a limit */
CD_SECS            =   60 #/* seconds per minute */
CD_FRAMES          =   75 #/* frames per second */
CD_SYNC_SIZE       =   12 #/* 12 sync bytes per raw data frame */
CD_MSF_OFFSET      =  150 #/* MSF numbering offset of first frame */
CD_CHUNK_SIZE      =   24 #/* lowest-level "data bytes piece" */
CD_NUM_OF_CHUNKS   =   98 #/* chunks per frame */
CD_FRAMESIZE_SUB   =   96 #/* subchannel data "frame" size */
CD_HEAD_SIZE       =    4 #/* header (address) bytes per raw data frame */
CD_SUBHEAD_SIZE    =    8 #/* subheader bytes per raw XA data frame */
CD_EDC_SIZE        =    4 #/* bytes EDC per most raw data frame types */
CD_ZERO_SIZE       =    8 #/* bytes zero per yellow book mode 1 frame */
CD_ECC_SIZE        =  276 #/* bytes ECC per most raw data frame types */
CD_FRAMESIZE       = 2048 #/* bytes per frame, "cooked" mode */
CD_FRAMESIZE_RAW   = 2352 #/* bytes per frame, "raw" mode */
CD_FRAMESIZE_RAWER = 2646 #/* The maximum possible returned bytes */
# most drives don't deliver everything:
CD_FRAMESIZE_RAW1  = (CD_FRAMESIZE_RAW-CD_SYNC_SIZE)  #  /*2340*/
CD_FRAMESIZE_RAW0  =(CD_FRAMESIZE_RAW-CD_SYNC_SIZE-CD_HEAD_SIZE) # /*2336*/

#typedef struct track
#{
#   unsigned short mode;
#   unsigned short minute;
#   unsigned short second;
#   unsigned short frame;
#
#   /* (min*60 + sec) * 75 + fps   */
#
#   unsigned long start_sector;
#
#   /* = the sizes in bytes off all tracks bevor this one */
#   /* its needed if there are mode1 tracks befor the mpeg tracks */
#   unsigned long start_offset;
#
#} tTrack;
track_struct = "HHHHLL"
# only 47 possible as http://mail.python.org/pipermail/python-checkins/2003-March/034383.html
tracks_size = struct.calcsize(track_struct) * 47

#struct virtualcd_info {
#   tTrack          tracks[100];
#   unsigned int    ffd;
#   int             num_tracks; /* number of tracks on the cd */
#  char            comment[256];
#};

cdemu_struct = "%dsIi255sx" % tracks_size

# default verbose is off
# 1 is on and 2 (currently used even more)
verbose = 0

class CDEmu_Error(StandardError):
    "our cdemu custom error class"

def parse_cue(cue_filename):
    "returns the supposable bin_filename and raw entries, if something went \
    wrong it throws an exception"
    # don't require an excact match .. so forget the ^ and $  ;-)
    regex_header = re.compile(r'FILE\s*"?(?P<bin_filename>.+?)"?\s*BINARY', re.IGNORECASE)
    regex_track = re.compile(r'TRACK\s*(?P<track>\d+)\s*(?P<mode>\S+)', re.IGNORECASE)
    regex_index = re.compile(r'INDEX\s*(?P<index>\d+)\s*(?P<msf>\d{1,2}\:\d{1,2}\:\d{1,2})', re.IGNORECASE)
    regex_ignore = re.compile(r'PREGAP\s|FLAGS\s', re.IGNORECASE)

    # read the file
    try:
        input = open (cue_filename,"r").readlines()
    except:
        raise CDEmu_Error, "Could not open the cue-file <%s>" % cue_filename
    
    # remove empty lines
    for i in range(len(input) - 1, 0):
        if not input[i].strip().rstrip('\n'):
            del(input[i])
    
    # the first line is the header .. match it
    found = regex_header.search(input[0])
    if found:
        if verbose: print "binfile: %s" % found.group("bin_filename")
        bin_filename = found.group("bin_filename")
    else:
        raise CDEmu_Error, "Invalid header of the cue-file <%s>" % cue_filename

    # remove the header
    del input[0]
    entries = []
 
    # look at the rest
    for line in input:
        found = regex_track.search(line)
        if found:
            if verbose: print "track: %s mode: %s" % (found.group("track"), found.group("mode"))
            entries.append([found.group("track"), found.group("mode"),[]])
            continue
    
        found = regex_index.search(line)
        if found:
            if verbose: print "index: %s msf: %s" % (found.group("index"), found.group("msf"))
            if entries != []:
                # insert my index data to the last track
                entries[-1][CUE_INDEXES].append([found.group("index"), found.group("msf")])
            else:
                if verbose: print "index before the first track --> ignoring"
            continue

        if regex_ignore.search(line):
            continue

        # we found a line we couldn't match ... thats an error
        print "Couldn't match: <%s>  (spaces striped)" % string.strip(line)
        
    return bin_filename, entries

def find_bin(supposable_bin_filename, cue_filename):
    "returns the fullpath of the bin file if found - the cue_filename must be with full path"

    # this will be a long long function, without any fun .. just trying
    # to find the fucking bin file
  
    # first try to find the file as given
    # this only works on good cue files ... does not happen
    # with "standard" windows generated cue files
    if os.path.isfile(supposable_bin_filename):
        return os.path.abspath(supposable_bin_filename)

    # now try it with the path of the cue file
    cue_path = os.path.dirname(cue_filename)
    s = os.path.basename(supposable_bin_filename)
    if verbose: print "Trying to open <%s>" % s
    if os.path.isfile(os.path.join(cue_path,s)): return os.path.join(cue_path,s)
    s = s.lower()
    if verbose: print "Trying to open <%s>" % s
    if os.path.isfile(os.path.join(cue_path,s)): return os.path.join(cue_path,s)
    s = s.upper()
    if verbose: print "Trying to open <%s>" % s
    if os.path.isfile(os.path.join(cue_path,s)): return os.path.join(cue_path,s)

    # now remove the path, try only the filename
    # try it as it is, lowercase, uppercase
    s = os.path.basename(supposable_bin_filename)
    if verbose: print "Trying to open <%s>" % s
    if os.path.isfile(s): return os.path.abspath(s)
    s = s.lower()
    if verbose: print "Trying to open <%s>" % s
    if os.path.isfile(s): return os.path.abspath(s)
    s = s.upper()
    if verbose: print "Trying to open <%s>" % s
    if os.path.isfile(s): return os.path.abspath(s)

    #remove the extension of the cuefile and add .bin/.img and try that
    s = os.path.splitext(os.path.basename(supposable_bin_filename))[0]
    if verbose: print "Trying to open <%s.bin>" % s
    if os.path.isfile(os.path.join(cue_path,s + ".bin")): return os.path.join(cue_path,s + ".bin")
    if verbose: print "Trying to open <%s.img>" % s
    if os.path.isfile(os.path.join(cue_path,s + ".img")): return os.path.join(cue_path,s + ".img")
  
    raise CDEmu_Error, "fuck .. giving up, tried 9 diff places, but couldn't find the bin-file <%s>" % supposable_bin_filename

def get_mode(mode):
    "converts the string to our internal mode consts"
    if mode.upper() == "AUDIO": return AUDIO
    if mode.upper() == "MODE1/2352": return MODE1_2352
    if mode.upper() == "MODE1/2048": return MODE1_2048
    if mode.upper() == "MODE2/2352": return MODE2_2352
    if mode.upper() == "MODE2/2336": return MODE2_2336
  
    raise CDEmu_Error, "Unkown mode <%s>" % mode

def mode_2_sectorsize(mode):
    if mode == AUDIO: return CD_FRAMESIZE_RAW
    if mode == MODE1_2352: return CD_FRAMESIZE_RAW
    if mode == MODE1_2048: return CD_FRAMESIZE
    if mode == MODE2_2352: return CD_FRAMESIZE_RAW
    if mode == MODE2_2336: return CD_FRAMESIZE_RAW0

    raise CDEmu_Error, "Unkown mode number:", mode

def get_msf(s):
    "converts 00:00:00 to 3 integers, take the 01 index if there is a choise - min, sec, frame"
    if len(s) > 1:
        for index in s:
            if int(index[0]) == 1:
                msf = index[1]
    else:
        msf = s[-1][1]
    min, sec, frame = msf.split(":")
    return int(min), int(sec), int(frame)

def msf_2_sector(min, sec, frame):
    "converts the msf to sector"
    return frame + (sec + min * 60 ) * 75;

def sector_2_msf(sector):
    "and the other way round - min,sec,frame"
    sector, frame = divmod(sector, 75)
    sector, sec = divmod(sector, 60)
    return sector, sec, frame

def extract_track_data(raw_entries, bin_filename):
    " generates out of the raw data, we got from the cue parser, useful data \
    we can use for our struct function"
    # first build an list we can calc with
  
    entries = []
  
    for i in range(0,len(raw_entries)):
        # mode / minute / second / frame
        mode = get_mode(raw_entries[i][CUE_MODE])
        min, sec, frame = get_msf(raw_entries[i][CUE_INDEXES])
        # svcds need 2 sec offset, don't really know why .. its just so ;-)
        if mode == MODE2_2352:
            sec+=2
        start_sector = msf_2_sector(min, sec, frame)
        
        if i == 0:
            start_offset = 0
        else:
            start_offset = entries[-1][ENTRY_STARTOFFSET] \
                + (start_sector -  entries[-1][ENTRY_STARTSECTOR]) * mode_2_sectorsize(entries[-1][ENTRY_MODE])

        entries.append([mode, min, sec, frame, start_sector, start_offset])
  
    # add the leadout
    mode = LEADOUT
    start_offset = os.path.getsize(bin_filename)
    start_sector = (start_offset - entries[-1][ENTRY_STARTOFFSET]) / mode_2_sectorsize(entries[-1][ENTRY_MODE]) + entries[-1][ENTRY_STARTSECTOR]
    min,sec,frame = sector_2_msf(start_sector)
    entries.append([mode, min, sec, frame, start_sector, start_offset])

    return entries

def build_struct(entries, bin_file_handle, num_tracks, comment):
    "generates the c struct we need"
    # loop over the entries
    tracks = ""
    for entry in entries:
        tracks+=struct.pack(track_struct,*entry)

    return struct.pack(cdemu_struct,tracks,bin_file_handle,num_tracks,comment)

def get_device_from_drive_number(devicenumber):
    if not devicenumber.isdigit():
        raise CDEmu_Error, "Device number is not a digit"

    # we'll try /dev/cdemu0 and ... 
    try:
        device = device_path + devicenumber
        if stat.S_ISBLK(os.stat(device)[stat.ST_MODE]):
            return device
    except:
        try:
            # /dev/cdemu/0
            device = device_path + '/' + devicenumber
            if stat.S_ISBLK(os.stat(device)[stat.ST_MODE]):
                return device    
        except:
            pass        
    # its no blockdevice .. argh
    raise CDEmu_Error, "could't find block device, if you've no devfs make sure you created them"
  
  
def load_cd(device_name, cue_filename, comment):
  
    # read the cue file
    cue_filename = os.path.abspath(cue_filename)
    supposable_bin_filename, raw_entries = parse_cue(cue_filename)
  
    # find the real bin-file
    bin_filename = find_bin(supposable_bin_filename, cue_filename)
    if verbose: print "real (existing) bin file: " , bin_filename

    # format the cue data into a way we need it
    tracks = extract_track_data(raw_entries, bin_filename)
    if verbose: print "reformated tracks: ", tracks

    # open the dev-device
    try:
        device = os.open(device_name, os.O_RDONLY | os.O_NONBLOCK)
        file = os.open(bin_filename,  os.O_RDONLY)
        # bild the c struct
        # -1 as we don't count the leadout
        c_struct = build_struct(tracks, file , len(tracks) - 1, comment)
  
        fcntl.ioctl(device, CDEMU_LOADCD, c_struct)
        os.close(device)
        
    except OSError, msg:
        raise CDEmu_Error, msg
        



def unload_cd(device_name):
    try:
        # open the dev-device
        device = os.open(device_name, os.O_RDONLY | os.O_NONBLOCK)
        fcntl.ioctl(device, CDEMU_UNLOADCD,0)
        os.close(device)
    except OSError, msg:
        raise CDEmu_Error, msg

if __name__ == '__main__':
    "only for testing this lib ... no use in the real life"
    # we need excatly one parameter
    if len(sys.argv) != 2:
        print "we need one parameter .. and that should be an filename ;-)"
        sys.exit()
  
    #load_cd("/dev/virtualcd0", sys.argv[1],"blabla")
    #unload_cd("/dev/virtualcd0")
    #print build_struct(tracks, 0, len(tracks), "blabla")
